// Controller.m - Charmap application delegate implementation
// Copyright (C) 2003 Christopher Culver <crculver@users.sourceforge.net>

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _CONTROLLER_H
#include "Controller.h"
#endif

@implementation Controller : NSObject
- (void) dealloc
{
  RELEASE(charHexInfo);
  RELEASE(charmapCell);
  RELEASE(charmapMatrix); 
  RELEASE(charmapWindow);
  RELEASE(largeDict);
  RELEASE(cellDict);
  RELEASE(display);
  RELEASE(inspectorImage);
  RELEASE(inspectorPanel);
  RELEASE(unicodeData);
  [super dealloc];
}

- (id) init
{
  NSMutableParagraphStyle *paragraphStyle=[NSMutableParagraphStyle defaultParagraphStyle];
  [paragraphStyle setAlignment: NSCenterTextAlignment];

  ASSIGN(cellDict, ([NSMutableDictionary dictionaryWithObjectsAndKeys:
			  [NSFont systemFontOfSize: 14],
			NSFontAttributeName,
			[NSColor blackColor],
			NSForegroundColorAttributeName,
			paragraphStyle,
			NSParagraphStyleAttributeName,
			nil]));
  ASSIGN(largeDict, ([NSMutableDictionary dictionaryWithObjectsAndKeys:
			  [NSFont systemFontOfSize: 48],
			NSFontAttributeName,
			[NSColor blueColor],
			NSForegroundColorAttributeName,
			paragraphStyle,
			NSParagraphStyleAttributeName,
			nil]));
  unicodeData = [UnicodeData new];
  inspectorOpen = NO;
  return self;
}

- (void) createMenu
{
  NSMenu *mainMenu;
  NSMenu *infoMenu;
  NSMenu *editMenu;
  NSMenu *viewMenu;

  mainMenu = [NSMenu new];

  // Info
  infoMenu = [NSMenu new];
  [infoMenu addItemWithTitle: @"Info Panel..."
	    action: @selector (orderFrontStandardInfoPanel:)
	    keyEquivalent: @""];
  [infoMenu addItemWithTitle: @"Help..."
	    action: @selector (orderFrontHelpPanel:)
	    keyEquivalent: @"?"];
  [mainMenu setSubmenu: infoMenu  forItem: [mainMenu addItemWithTitle: @"Info"
                                                               action: NULL
                                                        keyEquivalent: @""]];
  [infoMenu release];

  // Edit
  editMenu = [NSMenu new];
  [editMenu addItemWithTitle: @"Cut"
	    action: @selector (cut:)
	    keyEquivalent: @"x"];
  [editMenu addItemWithTitle: @"Copy"
	    action: @selector (copy:)
	    keyEquivalent: @"c"];
  [editMenu addItemWithTitle: @"Paste"
	    action: @selector (paste:)
	    keyEquivalent: @"v"];
  [editMenu addItemWithTitle: @"Select All"
	    action: @selector (selectAll:)
	    keyEquivalent: @"a"];
  [mainMenu setSubmenu: editMenu forItem: [mainMenu addItemWithTitle: @"Edit"
                                                              action: NULL
                                                       keyEquivalent: @""] ];
  RELEASE(editMenu);

  // View
  viewMenu = [[NSFontManager sharedFontManager] fontMenu: YES];
  [mainMenu setSubmenu: viewMenu forItem: 
    [mainMenu addItemWithTitle: @"View"
                        action: NULL
                 keyEquivalent: @""]];
  RELEASE(viewMenu);

  // Inspector 
  [mainMenu addItemWithTitle: @"Inspector"
	             action: @selector (showInspector:) 
	      keyEquivalent: @"I"];

  // Hide
  [mainMenu addItemWithTitle: @"Hide"
	    action: @selector (hide:)
	    keyEquivalent: @"h"];

  // Quit
  [mainMenu addItemWithTitle: @"Quit"
	action: @selector (terminate:)
	keyEquivalent: @"q"];

  [NSApp setMainMenu: mainMenu];
  RELEASE(mainMenu);
}

- (void) createWindow
{
  GSVbox *vbox;
  GSHbox *hbox;
  NSButton *copyButton;
  NSScrollView *charmapScroll;

  vbox = [GSVbox new];

  /*
   * Lower hbox will hold the text box and Copy button
  */
  hbox = [GSHbox new];
  [hbox setAutoresizingMask: NSViewWidthSizable];
  [hbox setBorder: 5];
  [hbox setDefaultMinXMargin: 75];
  display = [[NSTextField alloc] initWithFrame: NSMakeRect (0, 0, 432, 25)];
  [display setEditable: YES];
  [display setSelectable: YES];
  [display setAutoresizingMask: NSViewWidthSizable];
  [display setTarget: self];
  [display setAction: @selector (getSelected:)];
  [display setFont: [cellDict objectForKey: NSFontAttributeName]];
  [hbox addView: display];
  [display release];
  copyButton = [NSButton new];
  [copyButton setTitle: @"Copy"];
  [copyButton sizeToFit];
  [copyButton setTarget: self];
  [copyButton setAction: @selector (copySelected:)];
  [hbox addView: copyButton];
  [copyButton release];
  [vbox addView: hbox];
  [hbox release];
  
  // Upper hbox holds the blocks list and character map
  hbox = [GSHbox new];
  [hbox addSeparator];
  // Browser
  blocksBrowser = [NSBrowser new];
  [blocksBrowser setDelegate: unicodeData];
  [blocksBrowser setMaxVisibleColumns: 1];
  [blocksBrowser setFrameSize: NSMakeSize(250, 300)];
  [blocksBrowser setMinColumnWidth: 250];
  [blocksBrowser setAllowsMultipleSelection: NO];
  [blocksBrowser setAllowsEmptySelection: NO];
  [blocksBrowser setHasHorizontalScroller: NO];
  [blocksBrowser setTarget: self];
  [blocksBrowser setAction: @selector (generateCharmap)];
  [hbox addView: blocksBrowser];
  [blocksBrowser release];
  // Character map
  charmapScroll = [NSScrollView new];
  charmapCell = [[NSTextFieldCell alloc] init];
  charmapMatrix = [[NSMatrix alloc] initWithFrame: NSMakeRect (0, 0, 300, 300)
				    mode: NSRadioModeMatrix
				    prototype: charmapCell
				    numberOfRows: 0
				    numberOfColumns: 10];
  [charmapMatrix setAction: @selector (updateInspector:)];
  [charmapMatrix setDoubleAction: @selector (putSelected:)];
  [charmapMatrix setCellSize: NSMakeSize (30, 35)];
  [charmapMatrix setAllowsEmptySelection: YES];
  [charmapMatrix setAutoscroll: YES];
  [charmapScroll setDocumentView: charmapMatrix];
  [charmapScroll setFrameSize: NSMakeSize (330, 300)];
  [charmapScroll setHasVerticalScroller: YES];
  [hbox addView: charmapScroll];
  [charmapCell release];
  [charmapScroll release];
  [vbox addView: hbox];
  [hbox release];

  // Having packed everything in, display the window
  [vbox sizeToFit];
  winFrame.size = [vbox frame].size;
  winFrame.origin = NSMakePoint (0, 0);
  charmapWindow = [[NSWindow alloc] initWithContentRect: winFrame
				    styleMask: (NSTitledWindowMask
						| NSMiniaturizableWindowMask)
				    backing: NSBackingStoreBuffered
				    defer: YES];
  [charmapWindow setMinSize: winFrame.size];
  [charmapWindow setContentView: vbox];
  [vbox release];
  [charmapWindow setTitle: @"Charmap"];
  [charmapWindow center];
}

- (void) applicationWillFinishLaunching: (NSNotification *)not
{
  [self createMenu];
  [self createWindow];
  [self generateCharmap];
  [self createInspector];
}

- (void) applicationDidFinishLaunching: (NSNotification *)not;
{
  [charmapWindow orderFront: self];
}

- (void) generateCharmap
{
  NSString *tempString;
  NSTextFieldCell *tempCell;

  unichar aChar;
  int i, j, k, block, blockSize;
  unsigned int blockStart, blockEnd;
  
  // Remove all rows
  for (i = [charmapMatrix numberOfRows]; i >= 0; i--)
    {
      
      [charmapMatrix removeRow: i];
    }

  //  Add the necessary rows
  block = [blocksBrowser selectedRowInColumn: 0];
  if (block == -1)
    block = 0;
  blockSize = [unicodeData getBlockSize: block];
  blockStart = [unicodeData getBlockStart: block];
  blockEnd = [unicodeData getBlockEnd: block];
  for (i = 0; i < ((blockSize / 10) + 1); i++)
    {
      [charmapMatrix insertRow: 0];
    }

  // Fill the cells
  i = 0;
  for (k = blockStart; k < blockEnd; i++)
    {
      for (j = 0; j < 10; j++, k++)
	{
	  aChar = k;

	  tempString = [[NSString alloc] initWithCharacters: &aChar length: 1]; 
	  tempCell = [charmapMatrix cellAtRow: i column: j];
	  [tempCell setBackgroundColor: [NSColor whiteColor]];
	  [tempCell setDrawsBackground: YES];
	  [tempCell setAttributedStringValue: [[NSAttributedString alloc] initWithString: tempString
									  attributes: cellDict]];
	  [tempCell setAlignment: NSCenterTextAlignment];
	  [tempCell setTag: k];
	}
    }
  [tempString release];

  // Finish setting display
  [charmapMatrix sizeToCells];
  [charmapMatrix setNeedsDisplay: YES];
  [charmapMatrix scrollCellToVisibleAtRow: 1 column: 1];
  [charmapMatrix selectCellAtRow: 1 column: 1];
  if (inspectorOpen == YES)
    [self updateInspector: self];
}

- (void) putSelected: (id) sender
{
  NSMutableString *selectedString;
  int charTag;

  charTag = [[charmapMatrix selectedCell] tag];
  selectedString = [NSMutableString new];
  [selectedString appendString: [display stringValue]];
  [selectedString appendString: [[charmapMatrix selectedCell] stringValue]];
  [display setStringValue: selectedString];
}

- (void) copySelected: (id) sender
{
  NSPasteboard *pb = [NSPasteboard generalPasteboard];

  [display selectText: self];
  [pb declareTypes: [NSArray arrayWithObject: NSStringPboardType]
      owner: self];

  if ([[display stringValue] length] != 0)
    {
      [pb setString: [display stringValue] forType: NSStringPboardType];
      NSLog (@"Text copied\n");
    }
  else
    NSLog (@"No text selected.\n");
}

- (id) createInspector
{
  NSRect charNameInfoRect;
  GSHbox *hbox;
  GSVbox *inspectorView;

  inspectorView = [[GSVbox alloc] init];
  [inspectorView setAutoresizingMask: NSViewHeightSizable | NSViewWidthSizable];
  [inspectorView setBorder: 5];

  charHexInfo = [[NSTextField alloc] initWithFrame: NSMakeRect (0, 0, 1, 21)];
  [charHexInfo setAutoresizingMask: NSViewMaxXMargin];
  [charHexInfo setBezeled: NO];
  [charHexInfo setDrawsBackground: NO];
  [charHexInfo setEditable: NO];
  [charHexInfo setSelectable: NO];
  [inspectorView addView: charHexInfo];
  
  hbox = [[GSHbox alloc] init];
  [hbox setAutoresizingMask: NSViewWidthSizable];
  inspectorImage = [[NSTextView alloc] initWithFrame: NSMakeRect (0, 0,
								   75, 75)];

  [inspectorImage setAlignment: NSCenterTextAlignment];
  [inspectorImage setDrawsBackground: NO];
  [inspectorImage setEditable: NO];
  [inspectorImage setRichText: YES];
  [inspectorImage setSelectable: NO];
  [inspectorImage setTextColor: [NSColor blueColor]];
  [inspectorImage setUsesFontPanel: NO];
  [hbox addView: inspectorImage];

  charNameInfoRect = NSMakeRect(0, 0, 285, 50);
  charNameInfo = [[NSTextView alloc] initWithFrame: charNameInfoRect];
  [charNameInfo setAutoresizingMask: NSViewWidthSizable];
  [charNameInfo setBackgroundColor: [NSColor lightGrayColor]];
  [charNameInfo setEditable: NO];
  [charNameInfo setHorizontallyResizable: NO];
  [charNameInfo setSelectable: NO];
  [[charNameInfo textContainer] setContainerSize: NSMakeSize([charNameInfo frame].size.width, 50)];
  [[charNameInfo textContainer] setWidthTracksTextView: YES];
  [[charNameInfo textStorage] setAttributedString: [[NSAttributedString alloc]
						     initWithString: @"No character selected"]];
  [hbox addView: charNameInfo withMinXMargin: 15];
  [inspectorView addView: hbox];
  [hbox release];

  inspectorPanel = [[NSPanel alloc]
		     initWithContentRect: [inspectorView frame]
		     styleMask: NSTitledWindowMask
		     | NSClosableWindowMask
		     backing: NSBackingStoreBuffered
		     defer: YES];
  [inspectorPanel center];
  [inspectorPanel setContentView: inspectorView];
  [inspectorPanel setHidesOnDeactivate: YES];
  [inspectorPanel setMinSize: NSMakeSize (335,150)];
  [inspectorPanel setReleasedWhenClosed: NO];
  [inspectorPanel setTitle: @"Character Inspector"];
  [inspectorView release];
  return self;
}

- (id) updateInspector: (id) sender
{
  NSAttributedString *charNameAttrString;
  NSString *character;
  NSString *charNameString;
  int charTag;
  unichar chr;

  if (inspectorOpen == YES)
    {
      [[charmapMatrix selectedCell] setAlignment: NSCenterTextAlignment];
      
      charTag = [[charmapMatrix selectedCell] tag];
      chr = charTag;
      character = [[NSString alloc] initWithCharacters: &chr length: 1];
      charNameString = [unicodeData getName: charTag];
      charNameAttrString = [[NSAttributedString alloc] initWithString: charNameString];
      [charNameString release]; 
      
      [[inspectorImage textStorage] setAttributedString: [[NSAttributedString alloc] initWithString: character
										     attributes: largeDict]];
      
      [charHexInfo setStringValue: [[NSString stringWithFormat: @"U+%004x", charTag]
				     uppercaseString]];
      [charHexInfo sizeToFit];
      
      [[charNameInfo textStorage] setAttributedString: charNameAttrString];
    }
  return self;
}

- (id) showInspector: (id) sender
{
  inspectorOpen = YES;
  [self updateInspector: self];
  [inspectorPanel orderFront: self];
  return self;
}

- (void) changeFont: (id) sender
{
  NSFont *newCellFont;
  NSFont *newLargeFont;

  // Create new fonts
  newCellFont = [sender convertFont: [cellDict objectForKey: NSFontAttributeName]];
  [cellDict setObject: newCellFont forKey: NSFontAttributeName];

  newLargeFont = [sender convertFont: [largeDict objectForKey: NSFontAttributeName]];

  [largeDict setObject: [sender convertFont: newLargeFont toSize: 48] 
	     forKey: NSFontAttributeName];
  
  // Update display
  [display setFont: newCellFont];
  [self generateCharmap];
  [self updateInspector: self];
}

- (void) windowWillClose: (NSNotification *) theNotification
{
  inspectorOpen = NO;
}
@end
